﻿using Buddy.Coroutines;
using Clio.Utilities;
using ff14bot;
using ff14bot.Helpers;
using ff14bot.Managers;
using ff14bot.Navigation;
using ff14bot.Objects;
using System.Threading.Tasks;
using UltimaCR.Spells.Main;

namespace UltimaCR.Rotations
{
    public sealed partial class Summoner
    {
        private SummonerSpells _mySpells;

        private SummonerSpells MySpells
        {
            get { return _mySpells ?? (_mySpells = new SummonerSpells()); }
        }

        #region Class Spells

        private async Task<bool> MoveIntoRange()
        {
            if (Navigator.NavigationProvider != null)
            {
                if (Core.Me.CurrentTarget != null && (DataManager.GetSpellData(MySpells.Ruin.ID).Range < Core.Me.Distance(Core.Me.CurrentTarget) || !Core.Me.CurrentTarget.InLineOfSight()))
                {
                    if (!MovementManager.IsMoving && !MovementManager.IsTurning)
                    {
                        while (Core.Me.CurrentTarget != null && (DataManager.GetSpellData(MySpells.Ruin.ID).Range < Core.Me.Distance(Core.Me.CurrentTarget) || !Core.Me.CurrentTarget.InLineOfSight()))
                        {
                            Logging.Write(string.Format("[Ultima] Current distance to target is {0}, moving to within {1}", Core.Me.Distance(Core.Me.CurrentTarget), DataManager.GetSpellData(MySpells.Ruin.ID).Range));
                            MovementManager.SetFacing2D(Core.Me.CurrentTarget.Location);
                            await Coroutine.Wait(2000, () => Core.Me.IsFacing(Core.Me.CurrentTarget.Location));
                            if (!MovementManager.IsMoving) MovementManager.MoveForwardStart();
                            await Coroutine.Yield();
                        }
                        MovementManager.MoveForwardStop();
                    }
                }
            }

            return false;
        }

        private async Task<bool> YieldUnderXmilliseconds(int X)
        {
            if (!Core.Me.IsMounted)
            {
                while (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < X && DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds > 0)
                {
                    Logging.Write("[Ultima] Yielding while less then " + X + " milliseconds left on GCD");
                    await Coroutine.Yield();
                }
            }
            else Actionmanager.Dismount();
            return false;
        }

        private async Task<bool> Ruin()
        {
            if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 100)
            {
                await Coroutine.Yield();
            }
            if (MovementManager.IsMoving) { return false; }
            Logging.Write("[Ultima] Starting Ruin process");
            return await MySpells.Ruin.Cast();
        }

        private async Task<bool> Bio()
        {
            if (!Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true, 4000) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth / 4)
            {
                Logging.Write("[Ultima] Starting Bio process");
                return await MySpells.Bio.Cast();
            }
            return false;
        }

        private async Task<bool> Summon()
        {
            if (Core.Player.Pet == null &&
                Ultima.UltSettings.SummonerSummonPet &&
                Actionmanager.CanCast(MySpells.Summon.Name, Core.Player))
            {
                if (Ultima.UltSettings.SummonerGaruda ||
                    !Actionmanager.HasSpell(MySpells.SummonII.Name) &&
                    Ultima.UltSettings.SummonerTitan ||
                    !Actionmanager.HasSpell(MySpells.SummonIII.Name) &&
                    Ultima.UltSettings.SummonerIfrit)
                {
                    if (Ultima.UltSettings.SummonerSwiftcast
                        && Actionmanager.HasSpell(MySpells.CrossClass.Swiftcast.Name)
                        && Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Player))
                    {
                        Logging.Write("[Ultima] Starting Swiftcast_Summon process");
                        await MySpells.CrossClass.Swiftcast.Cast();
                        await Coroutine.Wait(1000, () => Core.Player.HasAura("Swiftcast"));
                    }
                    Logging.Write("[Ultima] Starting Summon process");
                    await MySpells.Summon.Cast();
                    await Coroutine.Wait(8000, () => Core.Player.Pet != null);
                    return Core.Player.Pet != null;
                }
            }
            return false;
        }

        private async Task<bool> Physick()
        {
            if (MovementManager.IsMoving) { return false; }
            if (Ultima.UltSettings.SummonerPhysick)
            {
                if (Core.Player.CurrentHealthPercent < 70)
                {
                    Logging.Write("[Ultima] Starting Physick process");
                    return await MySpells.Physick.Cast();
                }
            }
            return false;
        }

        private async Task<bool> Aetherflow()
        {
            if (!Core.Player.HasAura(MySpells.Aetherflow.Name) && Actionmanager.CanCast(MySpells.Aetherflow.Name, Core.Player))
            {
                if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500 && Core.Player.CurrentTarget != null && Core.Me.InCombat)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    Logging.Write("[Ultima] Starting RuinII_Aetherflow process");
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Aetherflow.Name, Core.Player));
                }
                Logging.Write("[Ultima] Starting Aetherflow process");
                return await MySpells.Aetherflow.Cast();
            }
            return false;
        }

        private async Task<bool> EnergyDrain()
        {
            if (Core.Player.HasAura(MySpells.Aetherflow.Name))
            {
                if ((Core.Player.CurrentManaPercent <= 90 &&
                    !Actionmanager.HasSpell(MySpells.Fester.Name)) ||
                    (Core.Player.CurrentManaPercent <= 40))
                {
                    if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                    {
                        await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                        Logging.Write("[Ultima] Starting RuinII_EnergyDrain process");
                        await MySpells.RuinII.Cast();
                        await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.EnergyDrain.Name, Core.Player.CurrentTarget));
                    }
                    Logging.Write("[Ultima] Starting EnergyDrain process");
                    return await MySpells.EnergyDrain.Cast();
                }
            }
            return false;
        }

        private async Task<bool> Miasma()
        {
            if (MovementManager.IsMoving) { return false; }
            if (!Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true, 5000) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth / 3)
            {
                Logging.Write("[Ultima] Starting Miasma process");
                return await MySpells.Miasma.Cast();
            }
            return false;
        }

        private async Task<bool> Virus()
        {
            Logging.Write("[Ultima] Starting Virus process");
            return await MySpells.Virus.Cast();
        }

        private async Task<bool> SummonII()
        {
            if (Core.Player.Pet == null &&
                Ultima.UltSettings.SummonerSummonPet &&
                Ultima.UltSettings.SummonerTitan &&
                Actionmanager.CanCast(MySpells.Summon.Name, Core.Player))
            {
                if (Ultima.UltSettings.SummonerSwiftcast
                    && Actionmanager.HasSpell(MySpells.CrossClass.Swiftcast.Name)
                    && Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Player))
                {
                    Logging.Write("[Ultima] Starting Swiftcast_SummonII process");
                    await MySpells.CrossClass.Swiftcast.Cast();
                    await Coroutine.Wait(1000, () => Core.Player.HasAura("Swiftcast"));
                }
                Logging.Write("[Ultima] Starting SummonII process");
                await MySpells.SummonII.Cast();
                await Coroutine.Wait(8000, () => Core.Player.Pet != null);
                return Core.Player.Pet != null;
            }
            return false;
        }

        private async Task<bool> Sustain()
        {
            if (MovementManager.IsMoving) { return false; }
            if (Core.Player.Pet != null &&
                Core.Player.Pet.CurrentHealthPercent < 70 &&
                !Core.Player.Pet.HasAura(MySpells.Sustain.Name))
            {
                Logging.Write("[Ultima] Starting Sustain process");
                return await MySpells.Sustain.Cast();
            }
            return false;
        }

        private async Task<bool> Resurrection()
        {
            if (MovementManager.IsMoving) { return false; }
            Logging.Write("[Ultima] Starting Resurrection process");
            return await MySpells.Resurrection.Cast();
        }

        private async Task<bool> BioII()
        {
            if (MovementManager.IsMoving) { return false; }
            if (!Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true, 5000) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth / 2)
            {
                Logging.Write("[Ultima] Starting BioII process");
                return await MySpells.BioII.Cast();
            }
            return false;
        }

        private async Task<bool> Bane()
        {
            if (Actionmanager.CanCast(MySpells.Bane.Name, Core.Me.CurrentTarget) && 
                Core.Player.HasAura(MySpells.Aetherflow.Name) &&
                Core.Player.CurrentManaPercent > 40 &&
                Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true, 8000) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true, 8000) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true, 5000))
            {
                if (Helpers.EnemiesNearTarget(8) >= 2)
                {
                    if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                    {
                        await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                        Logging.Write("[Ultima] Starting RuinII_Bane process");
                        await MySpells.RuinII.Cast();
                        await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Bane.Name, Core.Player.CurrentTarget));
                    }
                    Logging.Write("[Ultima] Starting Bane process");
                    return await MySpells.Bane.Cast();
                }
            }
            return false;
        }

        private async Task<bool> EyeForAnEye()
        {
            Logging.Write("[Ultima] Starting EyeForAnEye process");
            return await MySpells.EyeForAnEye.Cast();
        }

        private async Task<bool> RuinII()
        {
            if (MovementManager.IsMoving)
            {
                Logging.Write("[Ultima] Starting RuinII process");
                return await MySpells.RuinII.Cast();
            }
            if (Core.Player.CurrentManaPercent > 40)
            {
                /*if (Actionmanager.CanCast(MySpells.Aetherflow.Name, Core.Player) &&
                    !Core.Player.HasAura(MySpells.Aetherflow.Name) ||
                    Actionmanager.CanCast(MySpells.Fester.Name, Core.Player.CurrentTarget) &&
                    Helpers.EnemiesNearTarget(8) <= 1 &&
                    Ultima.UltSettings.SmartTarget ||
                    Actionmanager.CanCast(MySpells.Fester.Name, Core.Player.CurrentTarget) &&
                    Ultima.UltSettings.SingleTarget ||
                    Actionmanager.CanCast(MySpells.Bane.Name, Core.Player.CurrentTarget) &&
                    !Ultima.UltSettings.SingleTarget &&
                    Helpers.EnemiesNearTarget(8) > 1 &&
                    Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true, 8000) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true, 8000) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true, 5000) ||
                    Actionmanager.CanCast(MySpells.Rouse.Name, Core.Player) ||
                    Actionmanager.CanCast(MySpells.Spur.Name, Core.Player))
                {
                    return await MySpells.RuinII.Cast();
                }*/
                if (((Actionmanager.CanCast(MySpells.Aetherflow.Name, Core.Player) ||
                    Core.Me.CurrentManaPercent > 75) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true)) ||
                    Core.Me.CurrentManaPercent > 75) // && Core.Player.CurrentTarget.CurrentHealth < Core.Player.CurrentHealth/4)
                {
                    Logging.Write("[Ultima] Starting RuinII process");
                    return await MySpells.RuinII.Cast();
                }
            }
            return false;
        }

        private async Task<bool> Rouse()
        {
            if (Actionmanager.CanCast(MySpells.Rouse.Name, Core.Player) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth)
            {
                if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    Logging.Write("[Ultima] Starting RuinII_Rouse process");
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Rouse.Name, Core.Me));
                }
                Logging.Write("[Ultima] Starting Rouse process");
                return await MySpells.Rouse.Cast();
            }
            return false;
        }

        private async Task<bool> MiasmaII()
        {
            if (MovementManager.IsMoving) { return false; }
            if (Helpers.EnemiesNearPlayer(5) > 4 &&
                Core.Player.CurrentTarget.HasAura(MySpells.MiasmaII.Name, true, 3000))
            {
                Logging.Write("[Ultima] Starting MiasmaII process");
                return await MySpells.MiasmaII.Cast();
            }
            return false;
        }

        private static Vector3 shadowFlareLocation = new Vector3("0,0,0");
        private async Task<bool> ShadowFlare()
        {
            if (Core.Target != null &&
                Actionmanager.HasSpell(MySpells.ShadowFlare.Name) &&
                (!Core.Player.HasAura(MySpells.ShadowFlare.Name, true, 4000) ||
                Vector3.Distance(Core.Target.Location,shadowFlareLocation) > 5) &&
                Actionmanager.CanCastLocation(MySpells.ShadowFlare.Name, Core.Player.CurrentTarget.Location) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth)
            {
                Vector3 previousLocation = Core.Target.Location;
                await Coroutine.Wait(500, () => previousLocation != Core.Target.Location);
                if (previousLocation != Core.Target.Location) return false;
                /*if (Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget) &&
                    Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Player))
                {
                    if (await MySpells.RuinII.Cast())
                    {
                        await Coroutine.Wait(3000, () => Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Player));
                    }
                }
                if (Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Player))
                {
                    if (await MySpells.CrossClass.Swiftcast.Cast())
                    {
                        await Coroutine.Wait(3000, () => Actionmanager.CanCastLocation(MySpells.ShadowFlare.Name, Core.Player.CurrentTarget.Location) &&
                                                         Core.Player.HasAura(MySpells.CrossClass.Swiftcast.Name));
                    }
                }
                /*
                if (await MySpells.ShadowFlare.Cast())
                {
                    await Coroutine.Wait(3000, () => Core.Player.HasAura(MySpells.ShadowFlare.Name));
                }
                return true;*/
                shadowFlareLocation = Core.Target.Location;
                Logging.Write("[Ultima] Starting ShadowFlare process");
                return await MySpells.ShadowFlare.Cast();
            }
            return false;
        }

        #endregion

        #region Cross Class Spells

        #region Archer

        private async Task<bool> RagingStrikes()
        {
            if (Ultima.UltSettings.SummonerRagingStrikes)
            {
                if (Actionmanager.CanCast(MySpells.CrossClass.RagingStrikes.Name, Core.Player) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true) &&
                    Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth)
                {
                    if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                    {
                        await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                        Logging.Write("[Ultima] Starting RuinII_RagingStrikes process");
                        await MySpells.RuinII.Cast();
                        await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Rouse.Name, Core.Me));
                    }
                    Logging.Write("[Ultima] Starting RagingStrikes process");
                    return await MySpells.CrossClass.RagingStrikes.Cast();
                }
            }
            return false;
        }

        private async Task<bool> HawksEye()
        {
            if (Ultima.UltSettings.SummonerHawksEye)
            {
                Logging.Write("[Ultima] Starting HawksEye process");
                return await MySpells.CrossClass.HawksEye.Cast();
            }
            return false;
        }

        private async Task<bool> QuellingStrikes()
        {
            if (Ultima.UltSettings.SummonerQuellingStrikes)
            {
                Logging.Write("[Ultima] Starting QuellingStrikes process");
                return await MySpells.CrossClass.QuellingStrikes.Cast();
            }
            return false;
        }

        #endregion

        #region Thaumaturge

        private async Task<bool> Surecast()
        {
            if (Ultima.UltSettings.SummonerSurecast)
            {
                Logging.Write("[Ultima] Starting Surecast process");
                return await MySpells.CrossClass.Surecast.Cast();
            }
            return false;
        }

        private async Task<bool> BlizzardII()
        {
            if (MovementManager.IsMoving) { return false; }
            if (Ultima.UltSettings.SummonerBlizzardII)
            {
                Logging.Write("[Ultima] Starting BlizzardII process");
                return await MySpells.CrossClass.BlizzardII.Cast();
            }
            return false;
        }

        private async Task<bool> Swiftcast()
        {
            if (Ultima.UltSettings.SummonerSwiftcast)
            {
                Logging.Write("[Ultima] Starting Swiftcast process");
                return await MySpells.CrossClass.Swiftcast.Cast();
            }
            return false;
        }

        #endregion

        #endregion

        #region Job Spells

        private async Task<bool> Shockwave()
        {
            Logging.Write("[Ultima] Starting Shockwave process");
            return await MySpells.Shockwave.Cast();
        }

        private async Task<bool> MountainBuster()
        {
            Logging.Write("[Ultima] Starting MountainBuster process");
            return await MySpells.MountainBuster.Cast();
        }

        private async Task<bool> AerialSlash()
        {
            if (Core.Player.Pet != null &&
                Core.Player.Pet.Name == "Garuda-Egi")
            {
                /*if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.AerialSlash.Name, Core.Player.CurrentTarget));
                }*/
                Logging.Write("[Ultima] Starting AerialSlash process");
                return await MySpells.AerialSlash.Cast();
            }
            return false;
        }

        private async Task<bool> EarthenWard()
        {
            Logging.Write("[Ultima] Starting EarthenWard process");
            return await MySpells.EarthenWard.Cast();
        }

        private async Task<bool> SummonIII()
        {
            if (Core.Player.Pet == null &&
                Ultima.UltSettings.SummonerSummonPet &&
                Ultima.UltSettings.SummonerIfrit &&
                Actionmanager.CanCast(MySpells.Summon.Name, Core.Player))
            {
                if (Ultima.UltSettings.SummonerSwiftcast
                    && Actionmanager.HasSpell(MySpells.CrossClass.Swiftcast.Name)
                    && Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Player))
                {
                    Logging.Write("[Ultima] Starting Swiftcast_SummonIII process");
                    await MySpells.CrossClass.Swiftcast.Cast();
                    await Coroutine.Wait(1000, () => Core.Player.HasAura("Swiftcast"));
                }
                Logging.Write("[Ultima] Starting SummonIII process");
                await MySpells.SummonIII.Cast();
                await Coroutine.Wait(8000, () => Core.Player.Pet != null);
                return Core.Player.Pet != null;
            }
            return false;
        }

        public async Task<bool> SetPet()
        {
            /*if (Core.Player.Pet != null && Pet.PetStance != PetStance.Guard && !Core.Me.IsMounted)
            {
                if (Pet.PetStance != PetStance.Guard) Pet.DoAction("Guard", Core.Player);
                if (Core.Player.Pet.NpcId == 1404 || Core.Player.Pet.EnglishName == "Garuda - Egi")
                {
                    if (Pet.PetMode != PetMode.Obey || Core.Player.Pet.CurrentTargetId != Core.Me.CurrentTargetId) Pet.DoAction("Obey", Core.Player);
                }
                else if (Pet.PetMode != PetMode.Sic || Core.Player.Pet.CurrentTargetId != Core.Me.CurrentTargetId) Pet.DoAction("Sic", Core.Player);
            }
            else if(Core.Player.Pet != null && Pet.PetStance != PetStance.Steady && Core.Me.IsMounted)
            {
                if (Pet.PetStance != PetStance.Steady) Pet.DoAction("Steady", Core.Player);
            }*/

            if (Core.Player.Pet != null && !Core.Me.IsMounted)
            {
                if (Core.Me.CurrentTarget == null || Core.Me.CurrentTarget.CurrentHealth == 0)
                {
                    if (Pet.PetStance != PetStance.Steady)
                    {
                        Logging.Write("[Ultima] Starting Steady process");
                        Pet.DoAction("Steady", Core.Player);
                    }
                }
                else
                {
                    if (Pet.PetStance != PetStance.Guard)
                    {
                        Logging.Write("[Ultima] Starting Guard process");
                        Pet.DoAction("Guard", Core.Player);
                    }
                    if (Core.Player.Pet.NpcId == 1404 || Core.Player.Pet.EnglishName == "Garuda - Egi")
                    {
                        if (Pet.PetMode != PetMode.Obey || Core.Player.Pet.CurrentTargetId != Core.Me.CurrentTargetId)
                        {
                            Logging.Write("[Ultima] Starting Obey process");
                            Pet.DoAction("Obey", Core.Player);
                        }
                    }
                    else if (Pet.PetMode != PetMode.Sic || Core.Player.Pet.CurrentTargetId != Core.Me.CurrentTargetId)
                    {
                        Logging.Write("[Ultima] Starting Sic process");
                        Pet.DoAction("Sic", Core.Player);
                    }
                }
            }
            return false;
        }

        private async Task<bool> CrimsonCyclone()
        {
            return await MySpells.CrimsonCyclone.Cast();
        }

        private async Task<bool> RadiantShield()
        {
            return await MySpells.RadiantShield.Cast();
        }

        private async Task<bool> Contagion()
        {
            if (Core.Player.Pet != null &&
                Core.Player.Pet.Name == "Garuda-Egi" &&
                Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth)
            {
                if (Actionmanager.CanCast(MySpells.Contagion.Name, Core.Player.CurrentTarget) && (Core.Player.HasAura(MySpells.DreadwyrmTrance.Name)))
                {
                    /*if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                    {
                        await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                        await MySpells.RuinII.Cast();
                        await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Contagion.Name, Core.Me));
                    }*/
                    Logging.Write("[Ultima] Starting Contagion process");
                    return await MySpells.Contagion.Cast();
                }
            }
            return false;
        }

        private async Task<bool> Landslide()
        {
            Logging.Write("[Ultima] Starting Landslide process");
            return await MySpells.Landslide.Cast();
        }

        private async Task<bool> FlamingCrush()
        {
            Logging.Write("[Ultima] Starting FlamingCrush process");
            return await MySpells.FlamingCrush.Cast();
        }

        private async Task<bool> Fester()
        {
            if (Actionmanager.CanCast(MySpells.Fester.Name, Core.Me.CurrentTarget) && 
                Core.Player.HasAura(MySpells.Aetherflow.Name) &&
                Core.Player.CurrentManaPercent > 40 &&
                Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true))
            {
                if (Helpers.EnemiesNearTarget(8) <= 2)
                {
                    if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                    {
                        await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                        Logging.Write("[Ultima] Starting RuinII_Fester process");
                        await MySpells.RuinII.Cast();
                        await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Fester.Name, Core.Me));
                    }
                    Logging.Write("[Ultima] Starting Fester process");
                    return await MySpells.Fester.Cast();
                }
            }
            return false;
        }

        private async Task<bool> Tribind()
        {
            if (MovementManager.IsMoving) { return false; }
            Logging.Write("[Ultima] Starting Tribind process");
            return await MySpells.Tribind.Cast();
        }

        private async Task<bool> Spur()
        {
            if (Actionmanager.CanCast(MySpells.Spur.Name, Core.Player) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth)
            {
                if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    Logging.Write("[Ultima] Starting RuinII_Spur process");
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Spur.Name, Core.Me));
                }
                Logging.Write("[Ultima] Starting Spur process");
                return await MySpells.Spur.Cast();
            }
            return false;
        }

        private async Task<bool> Enkindle()
        {
            if (!Actionmanager.CanCast(MySpells.Rouse.Name, Core.Player) &&
                !Actionmanager.CanCast(MySpells.Spur.Name, Core.Player) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth)
            {
                if (Actionmanager.CanCast(MySpells.Enkindle.Name, Core.Player) && (Core.Player.HasAura(MySpells.DreadwyrmTrance.Name)))
                {
                    if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                    {
                        await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                        Logging.Write("[Ultima] Starting RuinII_Enkindle process");
                        await MySpells.RuinII.Cast();
                        await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Enkindle.Name, Core.Me));
                    }
                    Logging.Write("[Ultima] Starting Enkindle process");
                    return await MySpells.Enkindle.Cast();
                }
            }
            return false;
        }

        private async Task<bool> Painflare()
        {
            if (Actionmanager.CanCast(MySpells.Painflare.Name, Core.Player) &&
                Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true))
            {
                if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    Logging.Write("[Ultima] Starting RuinII_Painflare process");
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Painflare.Name, Core.Player.CurrentTarget));
                }
                Logging.Write("[Ultima] Starting Painflare process");
                return await MySpells.Painflare.Cast();
            }
            return false;
        }

        private async Task<bool> RuinIII()
        {
            if (MovementManager.IsMoving) { return false; }

            if (Actionmanager.CanCast(MySpells.RuinIII.Name, Core.Player.CurrentTarget) &&
                    (Core.Player.HasAura(MySpells.DreadwyrmTrance.Name) ||
                    (Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) &&
                    Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true) &&
                    Core.Me.CurrentManaPercent > 75)) ||
                    Core.Player.CurrentTarget.CurrentHealth < Core.Player.CurrentHealth / 4 &&
                    Core.Me.CurrentManaPercent > 90)
            {
                Logging.Write("[Ultima] Starting RuinIII process");
                return await MySpells.RuinIII.Cast();
            }
            return false;
        }

        private async Task<bool> SwiftRuinIII()
        {
            if (Actionmanager.CanCast(MySpells.RuinIII.Name, Core.Player.CurrentTarget) &&
                    Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Player) && 
                    Core.Player.HasAura(MySpells.DreadwyrmTrance.Name))
            {
                if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    Logging.Write("[Ultima] Starting RuinII_RuinIII process");
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.CrossClass.Swiftcast.Name, Core.Me));
                }
                Logging.Write("[Ultima] Starting Swiftcast_RuinIII process");
                if (await MySpells.CrossClass.Swiftcast.Cast())
                {
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.RuinIII.Name, Core.Player.CurrentTarget) && Core.Player.HasAura(MySpells.CrossClass.Swiftcast.Name));
                    Logging.Write("[Ultima] Starting RuinIII process");
                    return await MySpells.RuinIII.Cast();
                }
            }
            return false;
        }

        private async Task<bool> Tridisaster()
        {
            if (Actionmanager.CanCast(MySpells.Tridisaster.Name, Core.Player.CurrentTarget) && (Core.Player.HasAura(MySpells.DreadwyrmTrance.Name)))
            {
                if (!Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true) ||
                !Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) ||
                !Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true) &&
                Core.Player.CurrentTarget.CurrentHealth > Core.Player.CurrentHealth)
                {
                    if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                    {
                        await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                        Logging.Write("[Ultima] Starting RuinII_Tridisaster process");
                        await MySpells.RuinII.Cast();
                        await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Tridisaster.Name, Core.Player.CurrentTarget));
                    }
                    Logging.Write("[Ultima] Starting Tridisaster process");
                    return await MySpells.Tridisaster.Cast();
                }
            }
            return false;
        }

        private async Task<bool> DreadwyrmTrance()
        {
            if (Actionmanager.CanCast(MySpells.DreadwyrmTrance.Name, Core.Me) &&
                Core.Player.CurrentTarget.HasAura(MySpells.BioII.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Miasma.Name, true) &&
                Core.Player.CurrentTarget.HasAura(MySpells.Bio.Name, true) &&
                !Core.Player.HasAura(MySpells.DreadwyrmTrance.Name))
            {
                if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    Logging.Write("[Ultima] Starting RuinII_DreadwyrmTrance process");
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.DreadwyrmTrance.Name, Core.Me));
                }
                Logging.Write("[Ultima] Starting DreadwyrmTrance process");
                return await MySpells.DreadwyrmTrance.Cast();
            }
            return false;
        }

        private async Task<bool> Deathflare()
        {
            if (Actionmanager.CanCast(MySpells.Deathflare.Name, Core.Player.CurrentTarget) && (!Core.Player.HasAura(MySpells.DreadwyrmTrance.Name, true, 3000)) && (Core.Player.HasAura(MySpells.DreadwyrmTrance.Name)))
            {
                if (DataManager.GetSpellData(MySpells.RuinII.ID).Cooldown.TotalMilliseconds < 500)
                {
                    await Coroutine.Wait(1000, () => Actionmanager.CanCast(MySpells.RuinII.Name, Core.Player.CurrentTarget));
                    Logging.Write("[Ultima] Starting RuinII_Deathflare process");
                    await MySpells.RuinII.Cast();
                    await Coroutine.Wait(1500, () => Actionmanager.CanCast(MySpells.Deathflare.Name, Core.Me));
                }
                Logging.Write("[Ultima] Starting Deathflare process");
                return await MySpells.Deathflare.Cast();
            }
            return false;
        }

        #endregion

        #region PvP Spells

        private async Task<bool> AethericBurst()
        {
            return await MySpells.PvP.AethericBurst.Cast();
        }

        private async Task<bool> Equanimity()
        {
            return await MySpells.PvP.Equanimity.Cast();
        }

        private async Task<bool> ManaDraw()
        {
            return await MySpells.PvP.ManaDraw.Cast();
        }

        private async Task<bool> MistyVeil()
        {
            return await MySpells.PvP.MistyVeil.Cast();
        }

        private async Task<bool> Purify()
        {
            return await MySpells.PvP.Purify.Cast();
        }

        private async Task<bool> Wither()
        {
            return await MySpells.PvP.Wither.Cast();
        }

        #endregion
    }
}